# -*- coding: utf-8 -*-
"""
    vault-dump-restore
    ~~~~~~~~~~~~~~~~~~

    Dumps keys from a Hashicorp Vault instance to be able to restore it on
    another instance.
"""
__author__ = 'Hugo cisneiros'
__version__ = "0.4.1"



# -*- coding: utf-8 -*-
"""
    vault_dump_restore.cli
    ~~~~~~~~~~~~~~~~~~~~~~~

    This is the CLI interface for running the package.
"""
from argparse import ArgumentParser, RawDescriptionHelpFormatter
import logging
import os
import json
from vault_dump_restore import __version__
from vault_dump_restore.dump import VaultDumpKeys
from vault_dump_restore.restore import VaultRestoreKeys
import hvac

logger = logging.getLogger("vault_dump_restore")

def get_common_arguments():
    """
    Common arguments used by all commands.

    Each item from the list accepts a dict in this format::

        {
            "arg1,arg2": {
                "kwarg1": "value1",
                "kwarg2": "value2"
            }
        }

    Example::

        {
            "-p,--parameter": {
                "help": "This specifies a parameter in the example.",
                "dest": "parameter_var"
            }
        }

    Returns:
            list: A list of dicts containing the arguments to be used by
                method `argparse.ArgumentParser.add_parse()`
    """
    arguments_list = []
    arguments_list.append({
        "-d,--debug": {
            "help": 'Set LOG_LEVEL to DEBUG.',
            "dest": "debug_mode",
            "action": "store_true"
        }
    })

    arguments_list.append({
        "--address": {
            "help": f"Address of the Vault server. This can also be specified "
                f"via the VAULT_ADDR environment variable. "
                f"Defaults to https://127.0.0.1:8200",
            "dest": "vault_addr",
            "default": "https://127.0.0.1:8200"
        }
    })

    arguments_list.append({
        "--show-log-dates": {
            "help": 'Show log dates and severity on each message',
            "dest": "show_log_dates",
            "action": "store_true"
        }
    })

    arguments_list.append({
        "-p,--path": {
            "help": "Path to work on. This includes the secrets engine."
                "Example: secrets/subdir'. Defaults to /.",
            "dest": "path",
            "default": "/"
        }
    })

    return arguments_list

def common_setup(argument_parser):
    """
    Common checks and initial setup used by all commands

    Args:
        argument_parser: The parsed arguments returned from method
            `argparse.ArgumentParser.parse_args()`

    Returns:
        int: The status from the setup. 0 if everything went OK, >1 if not.
    """
    # logging setup
    if not logging.getLogger().hasHandlers():
        if argument_parser.show_log_dates:
            logformat='[%(asctime)s] [%(levelname)s] %(message)s'
        else:
            logformat='%(message)s'

        logging.basicConfig(format=logformat)

    logging.getLogger().setLevel(os.getenv("LOG_LEVEL", "FATAL"))

    # logging settings - environment variables are always a priority
    if argument_parser.debug_mode:
        logger.setLevel("DEBUG")
        logging.getLogger().setLevel(os.getenv("LOG_LEVEL", "DEBUG"))
    else:
        # defaults to info
        logger.setLevel("INFO")
        logging.getLogger().setLevel(os.getenv("LOG_LEVEL", "INFO"))

    # required environment variables check
    env_vars = os.environ.copy()
    if not "VAULT_TOKEN" in env_vars:
        logger.error("Error while reading environment variable VAULT_TOKEN. You need a valid/authenticated token.")
        return 1

    if not "VAULT_ADDR" in env_vars:
        os.environ["VAULT_ADDR"] = argument_parser.vault_addr

    # ensure utf-8 encoding
    os.environ["PYTHONIOENCODING"] = "utf-8"

    # vault authorization check
    hvac_client = hvac.Client(
        url=os.getenv("VAULT_ADDR"),
        token=os.getenv('VAULT_TOKEN')
    )

    if not hvac_client.is_authenticated():
        logger.error("Error while using the vault server using the token. Check its address and token.")
        return 1

    return 0

def dump():
    """
    Dump commnad called by CLI

    Returns:
        int: The status from the program. 0 if everything went OK, >1 if not.
    """
    # command argument parser
    parser = ArgumentParser(
        formatter_class=RawDescriptionHelpFormatter,
        description=(
            f"vault-dump {__version__}{os.linesep}"
            "Dumps keys from a Hashicorp Vault instance into a file"
        )
    )

    # load common arguments into the parser
    common_args = get_common_arguments()
    for argument in common_args:
        for args, kwargs in argument.items():
            args = args.split(",")
            parser.add_argument(*args, **kwargs)

    # dump options
    parser.add_argument(
        '-m',
        '--mask',
        help='All secrets in the dump will be hidden/masked.',
        dest="secrets_mask",
        action='store_true'
    )

    parser.add_argument(
        '-o',
        '--output',
        help='Choose output format. Could be: (json|vault). Defaults to json.',
        dest="output",
        default='json'
    )

    # parse the arguments, show in the screen if needed
    parser = parser.parse_args()

    # common setup
    setup_status = common_setup(parser)
    if setup_status > 0:
        return setup_status

    # check if output format is valid
    if not parser.output == "json" and not parser.output == "vault":
        logger.error(f"Invalid output format {parser.output}")

    dump = VaultDumpKeys(
        os.getenv("VAULT_ADDR"),
        os.getenv("VAULT_TOKEN"),
    )

    secrets_dump = dump.dump(parser.path, parser.secrets_mask)

    # JSON output
    if parser.output == "json":
        print(dump.dump_to_json(secrets_dump))

    # Vault client commands output
    elif parser.output == "vault":
        path_prefix = parser.path.lstrip("/")
        commands_dump = dump.dump_to_vault(secrets_dump, path_prefix)
        for command in commands_dump:
            print(command)

    return 0

def restore():
    """
    Restore command called by CLI

    Returns:
        int: The status from the program. 0 if everything went OK, >1 if not.
    """
    # command argument parser
    parser = ArgumentParser(
        formatter_class=RawDescriptionHelpFormatter,
        description=(
            f"vault-restore {__version__}{os.linesep}"
            "Restores keys from a file into a Hashicorp Vault instance"
        )
    )

    # load common arguments into the parser
    common_args = get_common_arguments()
    for argument in common_args:
        for args, kwargs in argument.items():
            args = args.split(",")
            parser.add_argument(*args, **kwargs)

    # restore options
    parser.add_argument('file', type=str,
        help='JSON file with the secrets')

    # parse the arguments, show in the screen if needed
    parser = parser.parse_args()

    # common setup
    setup_status = common_setup(parser)
    if setup_status > 0:
        return setup_status

    restore = VaultRestoreKeys(
        os.getenv("VAULT_ADDR"),
        os.getenv("VAULT_TOKEN"),
    )

    return 0



# -*- coding: utf-8 -*-
"""
    vault_dump_restore.restore
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    Restore methods
"""
import os
import json
import hvac

class VaultRestoreKeys:
    """
    Represents a keys restore to a vault instance
    """
    def __init__(self, hvac_url, hvac_token, skip_tls=False):
        """
        Creates a new restore object

        Args:
            hvac_url (str): Vault server to connect to
            hvac_token (str): Authentication token to be used when connecting
                to vault
            skip_tls (bool): Wheter to check for certificate or not while using
                HTTPS
        """
        self.hvac_client = hvac.Client(
            url=os.getenv(hvac_url),
            token=hvac_token
        )

    def restore(self, path):
        """
        Dumps a path recursively with all keys and values.

        Args:
            path (str): Path to look for keys.

        Returns:
            dict: all keys and values from the path
        """
        return {}


# -*- coding: utf-8 -*-
"""
    vault-dump-restore
    ~~~~~~~~~~~~~~~~~~

    Setup script for packaging and installing vault-dump-restore
"""
import pathlib
from setuptools import setup, find_packages

this_directory = pathlib.Path(__file__).parent.resolve()
long_description = (this_directory / 'README.md').read_text(encoding='utf-8')

setup(
    name="vault-dump-restore",
    version='0.4.1',
    description='Dumps keys from a Hashicorp Vault instance to be able to restore it on another instance',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Hugo Cisneiros (Eitch)',
    author_email='hugo.cisneiros@gmail.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3 :: Only',
    ],
    keywords='hashicorp, vault, keys, secrets, backup, restore',
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3.6, <4",
    install_requires=[
        'hvac==0.11.2'
    ],
    entry_points={
        'console_scripts': [
            'vault-dump=vault_dump_restore.cli:dump',
            'vault-restore=vault_dump_restore.cli:restore'
        ],
    },
)



# -*- coding: utf-8 -*-
"""
    vault_dump_restore.dump
    ~~~~~~~~~~~~~~~~~~~~~~~

    Dump methods
"""
import os
import json
import hvac

class VaultDumpKeys:
    """
    Represents a keys dump from a vault instance
    """
    def __init__(self, hvac_url, hvac_token, skip_tls=False):
        """
        Creates a new dump object

        Args:
            hvac_url (str): Vault server to connect to
            hvac_token (str): Authentication token to be used when connecting
                to vault
            skip_tls (bool): Wheter to check for certificate or not while using
                HTTPS
        """
        self.hvac_client = hvac.Client(
            url=os.getenv(hvac_url),
            token=hvac_token
        )
        self._sentinel = object()

    def dump(self, path, mask=False, root=False):
        """
        Dumps a path recursively with all keys and values.

        Args:
            path (str): Path to look for keys.
            mask (bool): If True, all secrets will be hidden in the values.
                Defaults to False.
            root (bool): If True, we treat the tree as the root of an engine,
                which uses a list and not a get.

        Returns:
            dict: all keys and values from the path
        """

        results = {}

        # when dumping the root
        if path == "/":
            # returns a list of secret engines
            engines = self.hvac_client.sys.list_mounted_secrets_engines()['data']
            for engine, values in engines.items():
                # we only accept KV2 engine types
                try:
                    if not values['type'] == "kv":
                        continue
                    elif values['options'] and not values['options']['version'] == "2":
                        continue
                except KeyError:
                    continue

                results[engine] = self.dump(path=engine, mask=mask, root=True)
        else:
            # split into the secret engine name and the actual secret path
            path_args = path.split("/")
            if root == True:
                # the root list of a secret engine
                path = "/"
                mount_point = path_args[0]
            else:
                # a sub-path of a secret engine
                path = "/".join(path_args[1:])
                mount_point = path_args[0]

            if path.endswith("/"):
                # this is a path, not a secret
                try:
                    listing = self.hvac_client.secrets.kv.v2.list_secrets(
                        path=path,
                        mount_point=mount_point
                    )['data']['keys']
                except hvac.exceptions.InvalidPath:
                    # we ignore empty paths
                    return results

                # recursively list keys within this path
                for key in listing:
                    results[key] = self.dump(path=f"{mount_point}/{path}{key}", mask=mask)
            else:
                # this is a secret, we read it
                try:
                    results = self.hvac_client.secrets.kv.v2.read_secret_version(
                        path=path,
                        mount_point=mount_point
                    )['data']['data']
                except hvac.exceptions.InvalidPath:
                    # we ignore destroyed secrets
                    return results

                # mask: we replace all values to a hidden value
                if mask:
                    for key in results:
                        results[key] = "<hidden>"

        return results

    def dump_to_json(self, dump, indent=2):
        """
        Transforms a dump into a JSON formatted output

        Args:
            dump (dict): Secrets dump to transform. This can be
                generated using the method dump.
            indent (int|Optional): Indentation to use on outputting
                the JSON. This parameter is passed to the json.dumps() method.
                None disables the pretty format.

        Returns:
            str: A JSON string with the dump contents.
        """
        return json.dumps(dump, indent=indent)

    def dump_to_vault(self, dump, path_prefix=""):
        """
        Transforms a dump into a series of vault client commands
        
        Args:
            dump (dict): Secrets dump to transform. This can be
                generated using the method dump.
            path_prefix (str|Optional): Prefix to be used before all lines.
                Defaults to an empty string.

        Returns:
            list: One command per item
        """
        # add trailing slash to path_prefix if needed
        if not path_prefix.endswith("/"):
            path_prefix = f"{path_prefix}/"

        lines = []
        old_key_path = None
        new_line = ""
        command_paths = self.dict_to_path(dump) 
        for is_last, dict_item in self._iter_check_last(command_paths):
            path, value = dict_item

            # separate key path from key itself
            key_path = "".join(path[:-1])
            key = path[-1]

            # add quotes if needed
            key = self._quotify_single(key)
            value = self._quotify_single(value)

            if key_path == old_key_path:
                # reuse the same key_path
                new_line = f"{new_line} '{key}'='{value}'"
            
            else:
                # first element is always new
                if not old_key_path:
                    new_line = f"{key_path}' '{key}'='{value}'"
                else:
                    # add the last line to the list
                    lines.append(f"vault kv put '{path_prefix}{new_line}")

                    # init a new key_path
                    new_line = f"{key_path}' '{key}'='{value}'"

            # last element
            if is_last:
                lines.append(f"vault kv put '{path_prefix}{new_line}")

            old_key_path = key_path

        return lines

    def dict_to_path(self, dictionary, path=None):
        """
        Transforms a dictionary's full path into one string. Means to be
        used recursivelly.

        Args:
            dictionary (dict): The dictionary itself
            path (str): Current path of the dictionary to traverse into
        """
        if path is None:
            path = []

        for key, value in dictionary.items():
            newpath = path + [key]

            if isinstance(value, dict):
                for u in self.dict_to_path(value, newpath):
                    yield u
            else:
                yield newpath, value

    def _iter_check_last(self, iterable):
        """
        Generator method that determines if an item is the last one during
        an iteration. Took from stackoverflow question 1630320, response
        by jsbueno.

        Args:
            iterable (mixed): An iterable object

        Yields:
            bool, item: True or False if it's the last item of the
                iterable, then the item from the iterable.
        """
        iterable = iter(iterable)
        current_element = next(iterable, self._sentinel)
        while current_element is not self._sentinel:
            next_element = next(iterable, self._sentinel)
            yield (next_element is self._sentinel, current_element)
            current_element = next_element

    def _quotify_single(self, string):
        """
        Adds bash quotes around a single quote. This is useful when a key
        or secret has the single quote character ('), which should be
        escaped.

        For example::
            mytesting'123 turns into: mytesting'"'"'123

        Args:
            string (str): String to be escaped

        Returns:
            string: Escaped string
        """
        return string.replace("'", "'\"'\"'")
